import pandas as _pd


def compute(cfg, inputDataFrame):
    return _pd.melt(inputDataFrame,
                    id_vars=[
                        c for c in inputDataFrame.columns if c not in cfg["cols"]
                    ],
                    value_vars=cfg["cols"],
                    var_name=cfg["col2"], value_name=cfg["col"])


def dimensions(cfg, inputDataFrame):
    return cfg["cols"]



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Parameterize(Component):
    """A Parameterize component.


Keyword arguments:

- id (string; required):
    The ID used to identify this component in Dash callbacks.

- config (boolean | number | string | dict | list; optional):
    The config the user sets in this component.

- meta (boolean | number | string | dict | list; required):
    The metadata this section is based on.

- meta_out (boolean | number | string | dict | list; optional):
    The metadata section will create as output."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, config=Component.UNDEFINED, meta=Component.REQUIRED, meta_out=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'config', 'meta', 'meta_out']
        self._type = 'Parameterize'
        self._namespace = 'dash_express_components'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'config', 'meta', 'meta_out']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in ['id', 'meta']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Parameterize, self).__init__(**args)



import plotly.graph_objects as _go
from scipy import stats as _stats
import pandas as _pd
import plotly.express as _px


def _get(inputDataFrame, plotConfigData):
    fig = _go.Figure()

    x = plotConfigData["params"]["x"]
    del plotConfigData["params"]["x"]

    showFit = False
    if "trendline" in plotConfigData["params"] and plotConfigData["params"]["trendline"] == "ols":
        showFit = True
        del plotConfigData["params"]["trendline"]

    groupParams = []
    # add color and facet entries to possible data groupings
    if "color" in plotConfigData["params"]:
        groupParams.append(plotConfigData["params"]["color"])
    if "facet_row" in plotConfigData["params"]:
        groupParams.append(
            plotConfigData["params"]["facet_row"])
    if "facet_col" in plotConfigData["params"]:
        groupParams.append(
            plotConfigData["params"]["facet_col"])
    groupParams = list(set(filter(None, groupParams)))

    def getProbPlot(groupData, grouped=True):
        res = _stats.probplot(groupData[x])
        if grouped:
            return _pd.Series({"Theoretical quantiles": res[0][0], "Percent": _stats.norm.cdf(res[0][0])*100.0, x: res[0][1], "Fit": res[0][0]*res[1][0] + res[1][1]})
        else:
            return _pd.DataFrame({"Theoretical quantiles": res[0][0], "Percent": _stats.norm.cdf(res[0][0])*100.0, x: res[0][1], "Fit": res[0][0]*res[1][0] + res[1][1]})

    if len(groupParams) > 0:
        plotData = inputDataFrame.groupby(groupParams).apply(
            lambda x: getProbPlot(x)).apply(_pd.Series.explode).reset_index()
    else:
        plotData = getProbPlot(inputDataFrame, False)

    # create the output plot
    
    plotConfigData["params"]["hover_data"] = ["Percent"]

    scatterPlot = _px.scatter(
        plotData, y="Theoretical quantiles", x=x,  **plotConfigData["params"])
    for d in scatterPlot.data:
        fig.add_trace(d)
    if showFit:
        for d in _px.line(plotData, y="Theoretical quantiles", x="Fit", **plotConfigData["params"]).data:
            d.showlegend = False
            fig.add_trace(d)
    fig.update_layout(scatterPlot.layout)
    fig.update_yaxes(
            tickmode='array',
            tickvals=[-3.090232306, -2.326347874, -1.281551566, -0.841621234, -0.524400513, -0.253347103,
                        0, 0.253347103, 0.524400513, 0.841621234, 1.281551566, 1.644853627, 2.326347874, 3.090232306],
            ticktext=["0.1%", "1%", "10%", "20%", "30%", "40%", "50%",
                        "60%", "70%", "80%", "90%", "95%", "99%", "99.9%"]
    )

    return fig


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class ConfigReceiver(Component):
    """A ConfigReceiver component.
A config receiver listening for `window.postMessage()`

@hideconstructor

@example
 rec = dxc.ConfigReceiver(
          id="plotConfig",
          token="test"
 )

 window.postMessage({config: "Test", token:"test" })
@public

Keyword arguments:

- id (string; required):
    The ID used to identify this component in Dash callbacks. @type
    {string}.

- config (boolean | number | string | dict | list; optional):
    Prop The resulting configuration of the plot. @type {Object}.

- token (string; required):
    A token used to define the configuration across frames. @type
    {string}."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, token=Component.REQUIRED, config=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'config', 'token']
        self._type = 'ConfigReceiver'
        self._namespace = 'dash_express_components'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'config', 'token']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in ['id', 'token']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(ConfigReceiver, self).__init__(**args)



import numpy as _np


def compute(cfg, inputDataFrame):

    # rename all columns with »
    trafo_cols = [c for c in inputDataFrame.columns if "»" in c]
    inputDataFrame.rename(
        columns={c: c.replace("»", "RightPointingDoubleAngle")
                 for c in trafo_cols},
        inplace=True
    )

    # rename all parts in the formula
    formula = cfg["formula"]
    for c in trafo_cols:
        new_c = c.replace("»", "RightPointingDoubleAngle")
        formula = formula.replace(c, new_c)

    var_dict = {
        "pi": _np.pi,
        "e": _np.e,
        "hbar": 6.58211951e-16,
        "c": 2.99792458e17,
        "hbar_c": 197.326979
    }
    inputDataFrame[cfg["col"]] = inputDataFrame.eval(
        formula, local_dict=var_dict, engine='numexpr')

    # make the back transformation
    inputDataFrame.rename(
        columns={c.replace("»", "RightPointingDoubleAngle"): c
                 for c in trafo_cols},
        inplace=True
    )

    return inputDataFrame


def dimensions(cfg, inputDataFrame):
    return [c for c in inputDataFrame.columns if c in cfg["formula"]]





def compute(cfg, inputDataFrame):
    inputDataFrame[cfg["col"]] = inputDataFrame[
        cfg["incol"]
    ].replace(cfg["values"])

    return inputDataFrame


def dimensions(cfg, inputDataFrame):
    return [cfg["incol"]]



from . import _aggr as aggr
from . import _eval as eval
from . import _combinecat  as combinecat
from . import _melt as melt 
from . import _catlookup as catlookup
from . import _dropna as dropna
from . import _wide_to_long as wide_to_long
from . import _zerostonan as zerostonan
from . import _rename as rename





def compute(cfg, inputDataFrame):
    inputDataFrame[cfg["col"]] = inputDataFrame[cfg["cols"]].apply(
        lambda row: '_'.join(row.values.astype(str)), axis=1)

    return inputDataFrame


def dimensions(cfg, inputDataFrame):
    return cfg["cols"]



import pandas as _pd
import plotly.express as _px


def _get(inputDataFrame, plotConfigData):

    # daw a count plot
    x = plotConfigData["params"]["x"]
    groupParams = [x]

    # add color and facet entries to possible data groupings
    if "color" in plotConfigData["params"]:
        groupParams.append(plotConfigData["params"]["color"])
    if "facet_row" in plotConfigData["params"]:
        groupParams.append(
            plotConfigData["params"]["facet_row"])
    if "facet_col" in plotConfigData["params"]:
        groupParams.append(
            plotConfigData["params"]["facet_col"])
    groupParams = list(set(filter(None, groupParams)))

    # group data and count the entries of each group
    plotData = inputDataFrame.groupby(groupParams)[x].count()
    # transoform the data back to have a count entry for each group tuple
    plotData = _pd.DataFrame(plotData)
    plotData.rename(columns={x: "count"}, inplace=True)
    plotData.reset_index(inplace=True)
    plotConfigData["params"]["y"] = "count"

    # plot the data
    return _px.bar(plotData, **plotConfigData["params"])


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Configurator(Component):
    """A Configurator component.
<div style="width:450px; margin-left: 20px; float: right;  margin-top: -150px;">
<img src="https://raw.githubusercontent.com/VK/dash-express-components/main/.media/configurator.png"/>
</div>


The configurator component helps to define plot definitions based on the
metadata of a dataframe.
Different configuration parts like `Filter`, `Transform` or `Plotter`
are combined in a single accordion component.

The metadata is used to compute the available parameters after data 
transformations and newly available colums are adjusted automatically.

@hideconstructor

@example
import dash_express_components as dxc
import plotly.express as px

meta = dxc.get_meta(px.data.gapminder())

 dxc.Configurator(
          id="plotConfig",
          meta=meta,
 )
@public

Keyword arguments:

- id (string; required):
    The ID used to identify this component in Dash callbacks. @type
    {string}.

- config (boolean | number | string | dict | list; optional):
    Prop The resulting configuration of the plot. @type {Object}.

- meta (boolean | number | string | dict | list; required):
    The metadata the plotter selection is based on. @type {Object}.

- showFilter (boolean; default True):
    Prop to define the visibility of the Filter panel @type {boolean}.

- showMetadata (boolean; default False):
    Prop to define the visibility of the Metadata panel @type
    {boolean}.

- showParameterization (boolean; default False):
    Prop to define the visibility of the Parameterization panel @type
    {boolean}.

- showPlotter (boolean; default True):
    Prop to define the visibility of the Plot panel @type {boolean}.

- showStore (boolean; default False):
    Prop to define the visibility of the Store panel @type {boolean}.

- showTransform (boolean; default True):
    Prop to define the visibility of the Transform panel @type
    {boolean}.

- showUpdate (boolean; default True):
    Prop to define the visibility of the update plot button @type
    {boolean}."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, meta=Component.REQUIRED, config=Component.UNDEFINED, showFilter=Component.UNDEFINED, showTransform=Component.UNDEFINED, showPlotter=Component.UNDEFINED, showMetadata=Component.UNDEFINED, showParameterization=Component.UNDEFINED, showStore=Component.UNDEFINED, showUpdate=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'config', 'meta', 'showFilter', 'showMetadata', 'showParameterization', 'showPlotter', 'showStore', 'showTransform', 'showUpdate']
        self._type = 'Configurator'
        self._namespace = 'dash_express_components'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'config', 'meta', 'showFilter', 'showMetadata', 'showParameterization', 'showPlotter', 'showStore', 'showTransform', 'showUpdate']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in ['id', 'meta']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Configurator, self).__init__(**args)



import pandas as _pd


def compute(cfg, inputDataFrame):
    output = _pd.wide_to_long(inputDataFrame,
                            stubnames=cfg["stubnames"],
                            i=cfg["i"],
                            j=cfg["j"],
                            sep=cfg["sep"],
                            suffix= {"string": "\w+", "number": "\d+"}[cfg["suffix"]]
                            ).reset_index()

    return output


def dimensions(cfg, inputDataFrame):
    return [
        *cfg["i"],
        *[c for c in inputDataFrame.columns
          if cfg["sep"] in c and cfg["sep"].join(c.split(cfg["sep"])[:-1]) in cfg["stubnames"]
          ]
    ]



import plotly.graph_objects as _go
import numpy as _np
import pandas as _pd
import plotly.express as _px


def _get(inputDataFrame, plotConfigData):
    fig = _go.Figure()
    if "x" in plotConfigData["params"]:
        # extrat params
        x = plotConfigData["params"]["x"]
        nbins = plotConfigData["params"]["nbins"]
        cumulative = True if "cumulative" in plotConfigData[
            "params"] and plotConfigData["params"]["cumulative"] else False
        histnorm = "count"
        normalize = False
        if "histnorm" in plotConfigData["params"]:
            if "density" in plotConfigData["params"]["histnorm"]:
                histnorm = "density"
                normalize = True
            if "percent" in plotConfigData["params"]["histnorm"]:
                histnorm = "percent"
                normalize = True

        groupParams = []
        # add color and facet entries to possible data groupings
        if "color" in plotConfigData["params"]:
            groupParams.append(
                plotConfigData["params"]["color"])
        if "facet_row" in plotConfigData["params"]:
            groupParams.append(
                plotConfigData["params"]["facet_row"])
        if "facet_col" in plotConfigData["params"]:
            groupParams.append(
                plotConfigData["params"]["facet_col"])
        groupParams = list(set(filter(None, groupParams)))


        plotData = _pd.DataFrame()
        if (len(groupParams) > 0):
            for group, allValues in inputDataFrame.groupby(groupParams):
                
                if not isinstance(group, list):
                    group = [group]

                # extract the sampels to compute the histogram
                histSamples = _np.array(
                    allValues[x].dropna().astype("float"))

                # compute the grid
                histX = _np.linspace(
                    _np.min(histSamples), _np.max(histSamples), nbins)
                gridNorm = histX[1]-histX[0]
                computeGrid = _np.linspace(
                    histX[0]-0.5*gridNorm, histX[-1]+0.5*gridNorm, nbins+1)

                # bin the data
                histY, _ = _np.histogram(
                    histSamples, bins=computeGrid, density=normalize)
                if cumulative:
                    histY = _np.cumsum(histY)
                if normalize:
                    histY = histY*gridNorm
                if histnorm == "percent":
                    histY *= 100.0

                colName = (
                    'cumulative ' if cumulative else '') + histnorm
                extra = _pd.DataFrame(
                    _np.transpose([histX, histY]))
                extra.columns = [x, colName]
                for gr_idx, gr_name in enumerate(groupParams):
                      extra[gr_name] = group[gr_idx]

                plotData = plotData.append(extra)

        else:
            # extract the sampels to compute the histogram
            histSamples = _np.array(inputDataFrame[x])

            # compute an grid
            histX = _np.linspace(
                _np.min(histSamples), _np.max(histSamples), nbins)
            gridNorm = histX[1]-histX[0]
            computeGrid = _np.linspace(
                histX[0]-0.5*gridNorm, histX[-1]+0.5*gridNorm, nbins+1)

            # bin the data
            histY, _ = _np.histogram(
                histSamples, bins=computeGrid, density=normalize)
            if cumulative:
                histY = _np.cumsum(histY)
            if normalize:
                histY = histY*gridNorm
            if histnorm == "percent":
                histY *= 100.0

            plotData = _pd.DataFrame(
                _np.transpose([histX, histY]))
            colName = (
                'cumulative ' if cumulative else '') + histnorm
            plotData.columns = [x, colName]

        params = {p: plotConfigData["params"][p] for p in plotConfigData["params"] if p not in [
            "nbins", "cumulative", "histnorm", "barmode"]}
        params["y"] = colName


        fig = _px.line(plotData, **params)

    return fig




def compute(cfg, inputDataFrame):
    inputDataFrame = inputDataFrame.dropna(
        subset=cfg["subset"]
    )

    return inputDataFrame


def dimensions(cfg, inputDataFrame):
    return cfg["subset"]



from plotly.subplots import make_subplots as _make_subplots
import plotly.graph_objects as _go
import numpy as _np


def _get(inputDataFrame, plotConfigData):

    fig = _go.Figure()

    extra_options = {}

    # if we have cols and rows, we want to make independent
    makeIndepX = False
    makeIndepY = False
    if "indep_x" in plotConfigData["params"]:
        makeIndepX = True
        del plotConfigData["params"]["indep_x"]
    if "indep_y" in plotConfigData["params"]:
        makeIndepY = True
        del plotConfigData["params"]["indep_y"]

    if "range_color" in plotConfigData["params"]:

        extra_options["zmin"] = plotConfigData["params"]["range_color"][0]
        extra_options["zmax"] = plotConfigData["params"]["range_color"][1]
        del plotConfigData["params"]["range_color"]

    if "x" in plotConfigData["params"] and "y" in plotConfigData["params"] and "dimensions" in plotConfigData["params"]:
        # extrat params
        x = plotConfigData["params"]["x"]
        y = plotConfigData["params"]["y"]
        dimensions = plotConfigData["params"]["dimensions"]

        params = {p: plotConfigData["params"][p] for p in plotConfigData["params"] if p in [
            "facet_col_wrap"]}
        params["binary_string"] = True
        params["contrast_rescaling"] = 'infer'
        params["color_continuous_scale"] = 'RdBu_r'

        extraGroupParams = []
        if "facet_col" in plotConfigData["params"]:
            extraGroupParams.append(
                plotConfigData["params"]["facet_col"])

        extraGroupParams = list(
            set(filter(None, extraGroupParams)))

        size = (
            int(inputDataFrame[y].max(
            ) - _np.minimum(0, inputDataFrame[y].min()))+1,
            int(inputDataFrame[x].max(
            ) - _np.minimum(0, inputDataFrame[x].min()))+1
        )
        all_min = (
            int(inputDataFrame[x].min()),
            int(inputDataFrame[y].min())
        )

        def _get_MAP_array(df, color):
            """internal helper to compue an array of the dataframe
            """
            try:
                v_df = df.pivot_table(
                    values=color, index=y, columns=x)

                # shuffle around, to fix all indexing problems :)
                output2 = _np.empty(
                    (size[0], len(v_df.columns.values)))
                output2[:, :] = _np.NaN
                output2[v_df.index.values.astype(
                    int)-all_min[0], :] = v_df.values

                output = _np.empty(size)
                output[:, :] = _np.NaN

                output[:, v_df.columns.values.astype(
                    int)-all_min[1]] = output2
                return output
            except Exception as ex:
                print(ex)
                return None

        if (len(extraGroupParams) > 0 or len(dimensions) > 1):
            data_dict = {}
            if len(extraGroupParams) > 0:
                for group, allValues in inputDataFrame.groupby(extraGroupParams):
                    for color in dimensions:
                        data_dict[f"{color} of {group}"] = _get_MAP_array(
                            allValues, color)
            else:
                for color in dimensions:
                    data_dict[f"{color}"] = _get_MAP_array(
                        inputDataFrame, color)

            import math
            col_wrap = 100000

            if "facet_col_wrap" in params:
                fig = _make_subplots(
                    math.ceil(len(data_dict) /
                              params["facet_col_wrap"]),
                    params["facet_col_wrap"],
                    shared_xaxes=not makeIndepX,
                    shared_yaxes=not makeIndepY,
                    subplot_titles=tuple(data_dict),
                )
                col_wrap = params["facet_col_wrap"]
            else:
                fig = _make_subplots(1, len(data_dict),
                                     shared_xaxes=not makeIndepX,
                                     shared_yaxes=not makeIndepY,
                                     subplot_titles=tuple(data_dict)
                                     )

            rowid = 1
            colid = 1
            for name, z in data_dict.items():
                fig.add_heatmap(
                    x0=all_min[0], dx=1,
                    y0=all_min[1], dy=1,
                    z=z,
                    name=name,
                    hovertemplate=f"{x}: %{{x}}<br>{y}: %{{y}}<br>{name.split(' of ')[0]}: %{{z}}<br>",
                    row=rowid,
                    col=colid,
                    coloraxis="coloraxis",
                    xaxis=None if makeIndepX else "x",
                    yaxis=None if makeIndepY else "y",
                )
                colid = colid + 1
                if colid > col_wrap:
                    rowid = rowid + 1
                    colid = 1

            if "zmin" in extra_options:
                fig.update_coloraxes(
                    cmin=extra_options["zmin"]
                )
            if "zmax" in extra_options:
                fig.update_coloraxes(
                    cmax=extra_options["zmax"]
                )

        else:
            fig.add_heatmap(
                x0=all_min[0], dx=1,
                y0=all_min[1], dy=1,
                z=_get_MAP_array(inputDataFrame, dimensions[0]),
                hovertemplate=f"{x}: %{{x}}<br>{y}: %{{y}}<br>{dimensions[0]}: %{{z}}<br>",
                **extra_options
            )

    return fig



from numpy import dtype
import pandas as _pd
import numpy as _np
from datetime import datetime, timedelta
from dateutil import parser

import plotly.express as _px
import plotly.graph_objects as _go
import base64 as _base64
import json as _json

from . import plottypes
from . import transformationtypes


_dummyData = _pd.DataFrame([None])


def get_meta(df):
    """
    extract the metadata from a dataframe needed to hand over to the Filter
    """

    def parse(key, val):
        if isinstance(val, _pd.CategoricalDtype):
            return {
                "type": "categorical",
                "cat": val.categories.tolist()
            }
        elif val == dtype('O'):
            return {
                "type": "categorical",
                "cat": df[key].unique().tolist()
            }
        elif val == dtype('bool'):
            return {
                "type": "bool"
            }
        elif "time" in str(val):
            return {"type": "temporal", **df[key].agg(["median", "min", "max"]).T.to_dict()}
        else:
            return {"type": "numerical", **df[key].agg(["median", "min", "max"]).T.to_dict()}

    return {
        k: parse(k, val)
        for k, val in df.dtypes.to_dict().items()
    }


def get_meta_dask(df):
    from numpy import dtype
    import dask
    """
    extract the metadata from a dask dataframe.
    """

    def parse(key, val):
        if val == dtype('O'):
            return {
                "type": "categorical",
                "cat": df[key].unique()
            }
        elif val == dtype('bool'):
            return {
                "type": "bool"
            }
        elif "time" in str(val):
            return {
                "type": "temporal",
                "min": df[key].min(),
                "max": df[key].max(),
                "median": df[key].mean()
            }
        else:
            return {
                "type": "numerical",
                "min": df[key].min(),
                "max": df[key].max(),
                "median": df[key].mean()
            }

    # create a dictionary of metadata to compute
    res = {
        k: parse(k, val)
        for k, val in df.dtypes.to_dict().items()
    }

    # compute it
    res = dask.compute(res, scheduler='single-threaded')[0]

    # transform the variables to simple python (json dump needed)
    for k, v in res.items():
        if "cat" in v:
            v["cat"] = v["cat"].tolist()
        for ik in ["min", "max", "median"]:
            if ik in v:
                v[ik] = float(v[ik])

    return res


def get_plot(inputDataFrame, config, apply_parameterization=True):

    errorResult = "Empty plot"

    try:
        # check if filter defined
        if config != None:

            # output figure
            fig = None

            configData = config
            # json parse the config if needed
            if isinstance(configData, str):
                configData = _json.loads(configData)

            plotConfigData = configData["plot"]
            # json parse the plotConfigconfig if needed
            if isinstance(plotConfigData, str):
                plotConfigData = _json.loads(plotConfigData)

            # don't try to hard if we have no params
            if not "params" in plotConfigData:
                return _px.scatter(_dummyData, title=errorResult)

            # also json parse the nested params, if needed
            if isinstance(plotConfigData["params"], str):
                plotConfigData["params"] = _json.loads(
                    plotConfigData["params"])

            # don't try too hard, if there is no plot axis
            if not("x" in plotConfigData["params"] or "y" in plotConfigData["params"] or "dimensions" in plotConfigData["params"]):
                return _px.scatter(_dummyData, title=errorResult)

            # apply filters if required
            if "filter" in configData:
                for el in configData["filter"]:
                    col = el["col"]
                    t = el["type"]
                    if t == "isin":
                        inputDataFrame = inputDataFrame[inputDataFrame[col].isin(
                            el["value"])]
                    elif t == "isnotin":
                        inputDataFrame = inputDataFrame[~inputDataFrame[col].isin(
                            el["value"])]
                    elif t == "gt":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]
                                                        > el["value"]]
                    elif t == "gte":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]
                                                        >= el["value"]]
                    elif t == "lt":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]
                                                        < el["value"]]
                    elif t == "lte":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]
                                                        <= el["value"]]
                    elif t == "eq":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]
                                                        == el["value"]]
                    elif t == "neq":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]
                                                        != el["value"]]
                    elif t == "istrue":
                        inputDataFrame = inputDataFrame[inputDataFrame[col]]
                    elif t == "isfalse":
                        inputDataFrame = inputDataFrame[~inputDataFrame[col]]
                    elif t == "after":
                        #df = df[df[col] > _pd.to_datetime(parser.isoparse(el["value"]))]
                        inputDataFrame = inputDataFrame[inputDataFrame[col] > _pd.Timestamp(
                            el["value"]).to_datetime64()]
                    elif t == "before":
                        #df = df[df[col] < _pd.to_datetime(parser.isoparse(el["value"]))]
                        inputDataFrame = inputDataFrame[inputDataFrame[col] < _pd.Timestamp(
                            el["value"]).to_datetime64()]
                    elif t == "lastn":
                        starttime = datetime.now() - \
                            timedelta(days=abs(el["value"]))
                        inputDataFrame = inputDataFrame[inputDataFrame[col] > starttime]

            # if dask mongodf or pandas:

            # compute usedCols
            usedCols = [b for a in [
                        (c if type(c) == list else [c]) for c in [
                            plotConfigData["params"].get(key) for key in
                            ["x", "y", "error_x", "error_y", "dimensions",
                                "color", "facet_col",
                                "size", "symbol", "pattern_shape",
                                "facet_row", "hover_name"]
                        ] if c is not None
                        ] for b in a]

            if "hover_data" in plotConfigData["params"]:
                usedCols.extend(plotConfigData["params"]["hover_data"])

            # add cols for transformations
            if "transform" in configData:
                for el in configData["transform"]:
                    usedCols.extend(
                        getattr(transformationtypes, el["type"]).dimensions(
                            el, inputDataFrame)
                    )

            usedCols = list(set(usedCols))

            # if we don't get a pandas dataframe it might be dask or mongodf
            if not isinstance(inputDataFrame, _pd.DataFrame):
                try:
                    inputDataFrame = inputDataFrame[[c for c in usedCols if c in inputDataFrame.columns]].compute(
                        scheduler='single-threaded'
                    )
                except Exception as err:
                    errorResult = "Error: " + str(err)
                    return _px.scatter(_dummyData, title=errorResult)
            else:
                inputDataFrame = inputDataFrame[[
                    c for c in usedCols if c in inputDataFrame.columns]].copy()

            # check if some data is left
            if len(inputDataFrame) == 0:
                return _px.scatter(_dummyData, title="No data available.")

            # apply data transformaitons if required
            if "transform" in configData:
                for el in configData["transform"]:
                    inputDataFrame = getattr(
                        transformationtypes, el["type"]).compute(el, inputDataFrame)

            # if we want to force an axis as categorical
            markCatX = False
            markCatY = False
            if "cat_x" in plotConfigData["params"]:
                markCatX = True
                del plotConfigData["params"]["cat_x"]
            if "cat_y" in plotConfigData["params"]:
                markCatY = True
                del plotConfigData["params"]["cat_y"]

            # if we have cols and rows, we want to make independent
            makeIndepX = False
            makeIndepY = False            
            if plotConfigData["type"] not in ["imshow"]:
                if "indep_x" in plotConfigData["params"]:
                    makeIndepX = True
                    del plotConfigData["params"]["indep_x"]
                if "indep_y" in plotConfigData["params"]:
                    makeIndepY = True
                    del plotConfigData["params"]["indep_y"]

            # if we have reversed x and y axis
            reversedX = False
            reversedY = False
            if "reversed_x" in plotConfigData["params"]:
                reversedX = True
                del plotConfigData["params"]["reversed_x"]
            if "reversed_y" in plotConfigData["params"]:
                reversedY = True
                del plotConfigData["params"]["reversed_y"]

            # add a sort command for categorical x columns
            if markCatX and "x" in plotConfigData["params"]:
                if not "category_orders" in plotConfigData["params"]:
                    plotConfigData["params"]["category_orders"] = {}
                val_name = plotConfigData["params"]["x"]
                cat_values = inputDataFrame[val_name].unique()
                cat_values.sort()
                plotConfigData["params"]["category_orders"][val_name] = cat_values.tolist(
                )

            # add a sort command for categorical x columns
            if markCatY and "y" in plotConfigData["params"]:
                if not "category_orders" in plotConfigData["params"]:
                    plotConfigData["params"]["category_orders"] = {}
                val_name = plotConfigData["params"]["y"]
                cat_values = inputDataFrame[val_name].unique()
                cat_values.sort()
                plotConfigData["params"]["category_orders"][val_name] = cat_values.tolist(
                )

            # add a sort command if multiple facets are present
            if "facet_row" in plotConfigData["params"]:
                if not "category_orders" in plotConfigData["params"]:
                    plotConfigData["params"]["category_orders"] = {}
                val_name = plotConfigData["params"]["facet_row"]
                cat_values = inputDataFrame[val_name].unique()
                cat_values.sort()
                plotConfigData["params"]["category_orders"][val_name] = cat_values.tolist(
                )

            if "facet_col" in plotConfigData["params"]:
                if not "category_orders" in plotConfigData["params"]:
                    plotConfigData["params"]["category_orders"] = {}
                val_name = plotConfigData["params"]["facet_col"]
                cat_values = inputDataFrame[val_name].unique()
                cat_values.sort()
                plotConfigData["params"]["category_orders"][val_name] = cat_values.tolist(
                )

            # remove nan values from dataframe
            # inputDataFrame = inputDataFrame.dropna(subset=[c for c in inputDataFrame.columns if c in usedCols])

            # return the corresponding plot
            if plotConfigData["type"] == "scatter":
                fig = _px.scatter(inputDataFrame, **
                                  plotConfigData["params"])
            if plotConfigData["type"] == "scatter_matrix":
                fig = _px.scatter_matrix(
                    inputDataFrame, **plotConfigData["params"])
            if plotConfigData["type"] == "density_heatmap":
                fig = _px.density_heatmap(
                    inputDataFrame, **plotConfigData["params"])
            if plotConfigData["type"] == "density_contour":
                fig = _px.density_contour(
                    inputDataFrame, **plotConfigData["params"])
            if plotConfigData["type"] == "violin":
                fig = _px.violin(inputDataFrame, **
                                 plotConfigData["params"])
            if plotConfigData["type"] == "bar":
                if "x" in plotConfigData["params"] and "y" in plotConfigData["params"]:
                    fig = _px.bar(inputDataFrame, **
                                  plotConfigData["params"])
            if plotConfigData["type"] == "histogram":
                if "x" in plotConfigData["params"]:
                    fig = _px.histogram(
                        inputDataFrame, **plotConfigData["params"])

            # call extra plotters
            if hasattr(plottypes,  plotConfigData["type"]):
                fig = getattr(plottypes, plotConfigData["type"])(
                    inputDataFrame, plotConfigData)

            # if we want to force an axis as categorical
            if "x" in plotConfigData["params"] and markCatX:
                fig.update_xaxes(type='category')
            if "y" in plotConfigData["params"] and markCatY:
                fig.update_yaxes(type='category')

            if "title" in plotConfigData and plotConfigData["title"] is not None and plotConfigData["title"] != "":
                fig.update_layout(title=plotConfigData["title"])

            # if we want to have independent x and y axis
            if makeIndepX:
                fig.update_xaxes(matches=None, showticklabels=True)
                for idx, a in enumerate([a for a in fig.layout if "xaxis" in a]):
                    d_min, d_max = fig.layout[a].domain
                    d_center = 0.5*(d_min+d_max)
                    d_width = 0.43*(d_max - d_min)
                    fig.layout[a].domain = (
                        d_center - d_width, d_center+d_width)
            if makeIndepY:
                fig.update_yaxes(matches=None, showticklabels=True)

            # reverse the axis
            if reversedX:
                fig.update_xaxes(autorange="reversed")
            if reversedY:
                fig.update_yaxes(autorange="reversed")                

            # transform the fig to an png if object too big
            if "render" in plotConfigData and "png" in plotConfigData["render"]:

                try:
                    # prepare to compute the image on a remote orca server
                    fig.update_layout(
                        template="plotly_white"
                    )
                    img_bytes = fig.to_image(
                        format="png", width=1200, height=700, scale=2)
                    encoded = _base64.b64encode(img_bytes)

                    # create an empty figure with some fixed axis
                    fig = _go.Figure()
                    fig.update_layout(
                        xaxis=_go.layout.XAxis(
                            showticklabels=False,
                            showgrid=False,
                            zeroline=False,
                            range=[0, 1200]
                        ),
                        yaxis=_go.layout.YAxis(
                            showticklabels=False,
                            showgrid=False,
                            zeroline=False,
                            range=[0, 700],
                            scaleanchor='x'
                        ),
                        margin={'l': 0, 'r': 0, 't': 0, 'b': 0},
                        autosize=True,
                    )

                    # add the png to the figure
                    fig.add_layout_image(
                        dict(
                            source="data:image/png;base64,{}".format(
                                encoded.decode('ascii')),
                            xref="x",
                            yref="y",
                            x=0,
                            sizex=1200,
                            y=700,
                            sizey=700,
                            opacity=1.0,
                            layer="below",
                            sizing="stretch"
                        )
                    )

                except Exception as ex:
                    print(ex)

            return fig

    except Exception as inst:
        errorResult = "Error: " + str(inst)

    return _px.scatter(_dummyData, title=errorResult)



import plotly.express as _px


def q01(x):
    return x.quantile(0.01)


def q05(x):
    return x.quantile(0.05)


def q25(x):
    return x.quantile(0.25)


def q75(x):
    return x.quantile(0.75)


def q95(x):
    return x.quantile(0.95)


def q99(x):
    return x.quantile(0.99)


def iqr(x):
    return x.quantile(0.75) - x.quantile(0.25)


def range(x):
    return x.max() - x.min()


aggr_func = {
    "q01": q01,
    "q05": q05,
    "q25": q25,
    "q75": q75,
    "q95": q95,
    "q99": q99,
    "iqr": iqr,
    "range": range
}


def _get_dict(x, y, text):
    return dict(showarrow=False,
                x=x,
                y=y,
                text=text,
                font_size=10,
                bgcolor="#ffffff",
                opacity=0.6
                )


def _get(inputDataFrame, plotConfigData):

    params = plotConfigData["params"]
    aggr_groups = {}
    aggr_needed = []
    if "aggr" in params:
        # check if aggr is needed
        aggr = params["aggr"]
        if isinstance(aggr, str):
            aggr = [aggr]
        del params["aggr"]

        for el in aggr:
            if el in ["mean", "median", "count", "range", "iqr"]:
                aggr_needed.append("median")
                aggr_needed.append(el)
                if "median" not in aggr_groups:
                    aggr_groups["median"] = {}
                if el == "count":
                    aggr_groups["median"][el] = {
                        "func": lambda k, v: f"<b>#</b> {int(v)}"
                    }
                else:
                    aggr_groups["median"][el] = {
                        "func": lambda k, v: f"<b>{k}</b><br>{v:.3E}"
                    }

            else:
                aggr_needed.append(el)
                aggr_groups[el] = {}
                aggr_groups[el][el] = {
                    "func": lambda k, v: f"<b>{k}</b><br>{v:.3E}"
                }
        aggr_needed = list(set(aggr_needed))
        aggr_needed = [aggr_func[t]
                       if t in aggr_func else t for t in aggr_needed]

    fig = _px.box(inputDataFrame, **params)


    if "x" in params and "y" in params and not "facet_col" in params and not "facet_row" in params:
        for idx, val in inputDataFrame.groupby(params["x"])[params["y"]].aggregate(aggr_needed).iterrows():

            for gr_name, gr in aggr_groups.items():

                add_str_array = [
                    v["func"](k, val[k])
                    for k, v in gr.items()
                ]
                fig.layout.annotations = [
                    *fig.layout.annotations,
                    _get_dict(
                        idx, val[gr_name], "<br>".join(add_str_array))
                ]

    if not "x" in params and "y" in params and not "facet_col" in params and not "facet_row" in params:
        val = inputDataFrame[params["y"]].aggregate(aggr_needed)
        for gr_name, gr in aggr_groups.items():

            add_str_array = [
                v["func"](k, val[k])
                for k, v in gr.items()
            ]
            fig.layout.annotations = [
                *fig.layout.annotations,
                _get_dict(
                    0, val[gr_name], "<br>".join(add_str_array))
            ]

    if "x" in params and not "y" in params and not "facet_col" in params and not "facet_row" in params:
        val = inputDataFrame[params["x"]].aggregate(aggr_needed)
        for gr_name, gr in aggr_groups.items():
            add_str_array = [
                v["func"](k, val[k])
                for k, v in gr.items()
            ]
            fig.layout.annotations = [
                *fig.layout.annotations,
                _get_dict(
                    val[gr_name], 0, "<br>".join(add_str_array))
            ]

    return fig



import numpy as np

def compute(cfg, inputDataFrame):

    for c in cfg["subset"]:
        inputDataFrame[c].replace(0.0, np.nan, inplace=True)

    return inputDataFrame


def dimensions(cfg, inputDataFrame):
    return cfg["subset"]



from .ConfigReceiver import ConfigReceiver
from .Configurator import Configurator
from .Filter import Filter
from .Graph import Graph
from .Localstore import Localstore
from .MetaCheck import MetaCheck
from .Parameterize import Parameterize
from .Plotter import Plotter
from .Transform import Transform

__all__ = [
    "ConfigReceiver",
    "Configurator",
    "Filter",
    "Graph",
    "Localstore",
    "MetaCheck",
    "Parameterize",
    "Plotter",
    "Transform"
]


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Filter(Component):
    """A Filter component.
<div style="width:450px; margin-left: 20px; float: right;  margin-top: -150px;">
<img src="https://raw.githubusercontent.com/VK/dash-express-components/main/.media/filter.png"/>
<img src="https://raw.githubusercontent.com/VK/dash-express-components/main/.media/filter-modal.png"/>
</div>


The filter component allows filters to be set on individual columns.

These are all combined with and.

There are the options like >, <, = for numerical, as well as isin etc. for categorical columns.
Possible options are given by the `meta` props.


@hideconstructor

@example
import dash_express_components as dxc
import plotly.express as px

meta = dxc.get_meta(px.data.gapminder())

dxc.Filter(
  id="filter",
  meta=meta
)
@public

Keyword arguments:

- id (string; required):
    The ID used to identify this component in Dash callbacks.

- config (boolean | number | string | dict | list; optional):
    The config the user sets in this component.

- meta (boolean | number | string | dict | list; required):
    The metadata this section is based on.

- meta_out (boolean | number | string | dict | list; optional):
    The metadata section will create as output."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, config=Component.UNDEFINED, meta=Component.REQUIRED, meta_out=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'config', 'meta', 'meta_out']
        self._type = 'Filter'
        self._namespace = 'dash_express_components'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'config', 'meta', 'meta_out']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in ['id', 'meta']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Filter, self).__init__(**args)



import json
import os
from setuptools import setup, find_packages



with open('package.json') as f:
    package = json.load(f)

package_name = package["name"].replace(" ", "_").replace("-", "_")

setup(
    name=package_name,
    version=package["version"],
    author=package['author'],
    packages=[p for p in find_packages() if p.startswith(package_name)],
    include_package_data=True,
    license=package['license'],
    description=package.get('description', package_name),
    install_requires=["numpy", "pandas", "numexpr"],
    classifiers = [
        'Framework :: Dash',
    ],    
)


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Graph(Component):
    """A Graph component.
<div style="width:450px; margin-left: 20px; float: right;  margin-top: -150px;">
<img src="https://raw.githubusercontent.com/VK/dash-express-components/main/.media/graph.png"/>
<img src="https://raw.githubusercontent.com/VK/dash-express-components/main/.media/graph-table.png"/>
<img src="https://raw.githubusercontent.com/VK/dash-express-components/main/.media/graph-modal.png"/>
</div>


The `Graph` component is a combination of the original dash `Graph` and the dash `data_table`.

It can not only be used to render a plotly.js-powered data visualization,
but also shows a searchable table, if only data is submitted.

In addition, there is the possibility to add plot parameters as `defParams` and 
the dataframe `meta` data.  
This automatically adds a configurator modal, which can be opened via a button
at the bottom right.


@hideconstructor

@example
import dash_express_components as dxc
import plotly.express as px

meta = dxc.get_meta(px.data.gapminder())

dxc.Graph(
    id="fig",
    meta=meta,
    defParams={}
)
@public

Keyword arguments:

- id (string; required):
    The ID of this component, used to identify dash components in
    callbacks. The ID needs to be unique across all of the components
    in an app. @type {string}.

- className (string; optional):
    className of the parent div.

- defParams (dict; optional):
    Configuration to describe the plot features.

- figure (boolean | number | string | dict | list; default {    data: [],    layout: {},    frames: [],}):
    Plotly `figure` object. See schema:
    https://plotly.com/javascript/reference  `config` is set
    separately by the `config` property.

- meta (boolean | number | string | dict | list; optional):
    The metadata the plotter selection is based on.

- selectedData (boolean | number | string | dict | list; optional):
    The data selected in the plot or in the table.

- style (dict; optional):
    Generic style overrides on the plot div."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, defParams=Component.UNDEFINED, meta=Component.UNDEFINED, figure=Component.UNDEFINED, style=Component.UNDEFINED, selectedData=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'className', 'defParams', 'figure', 'meta', 'selectedData', 'style']
        self._type = 'Graph'
        self._namespace = 'dash_express_components'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'className', 'defParams', 'figure', 'meta', 'selectedData', 'style']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in ['id']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Graph, self).__init__(**args)



def q01(x):
    return x.quantile(0.01)


def q05(x):
    return x.quantile(0.05)


def q25(x):
    return x.quantile(0.25)


def q75(x):
    return x.quantile(0.75)


def q95(x):
    return x.quantile(0.95)


def q99(x):
    return x.quantile(0.99)


def iqr(x):
    return x.quantile(0.75) - x.quantile(0.25)


def range(x):
    return x.max() - x.min()


aggr_func = {
    "q01": q01,
    "q05": q05,
    "q25": q25,
    "q75": q75,
    "q95": q95,
    "q99": q99,
    "iqr": iqr,
    "range": range
}


def compute(cfg, inputDataFrame):

    types = [aggr_func[t] if t in aggr_func else t for t in cfg["types"]]

    output = inputDataFrame.groupby(
        cfg["groupby"])[cfg["cols"]].agg(types).reset_index()
    output.columns = ["_".join([cc for cc in c if cc and cc != ""])
                      for c in output.columns]

    return output


def dimensions(cfg, inputDataFrame):
    return [
        *cfg["groupby"],
        *cfg["cols"]
    ]



from ._bar_count import _get as bar_count
from ._box import _get as box
from ._histogram_line import _get as histogram_line
from ._imshow import _get as imshow
from ._probability import _get as probability
from ._table import _get as table

